import copy
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
from tiatoolbox.tools import stainnorm


class TransformerAdapter:
    """
    Parent class for all Adapters. Subclasses need to implement their call method.
    """
    def __init__(self):
        self.is_callable_with_batches = False

    def __call__(self, img_or_imgs): # TODO specify type, as well for subclasses
        raise NotImplementedError


class RandomPrinterAdapter(TransformerAdapter):
    """
    TODO
    """
    def __init__(self, message: str = "RandomPrinterAdapter:"):
        """
        TODO

        Arguments:
            message: TODO
        """

        super().__init__()
        self.message = message
        self.is_callable_with_batches = True

    def __call__(self, img_or_imgs): # TODO specify type
        """
        TODO

        Arguments:
            img_or_imgs: TODO

        Returns:
            TODO: TODO
        """
        print(self.message, np.random.random())
        return img_or_imgs


class CopyAdapter(TransformerAdapter):
    """
    CopyAdapter creating a shallow copy of the input
    """
    def __init__(self, num_of_copies: int = 2):
        """
        TODO

        Arguments:
            num_of_copies: TODO
        """

        super().__init__()
        self.is_callable_with_batches = True
        self.num_of_copies = num_of_copies

    def __call__(self, img_or_imgs) -> dict: # TODO specify type
        """
        TODO

        Arguments:
            img_or_imgs: TODO
            
        Returns:
            TODO: TODO
        """

        result = {}
        for i in range(self.num_of_copies):
            # TODO
            # result.append(img_or_imgs.clone())
            result[str(i)] = copy.copy(img_or_imgs)
        return result


def convert_to_uint8_image(img: Any): # TODO Could either use extra package for typing or wait till we can upgrade to 3.8+: https://numpy.org/devdocs/reference/typing.html#module-numpy.typing
    """
    Convert any image to an unsigned int.
    
    Arguments:
        img: TODO        
            
    Returns:
        TODO: TODO
    """
    
    if img.dtype != np.uint8:
        img = img * 255 + 0.5
        img[img > 255] = 255
        return np.array(img, dtype=np.uint8)
    else:
        return img


class TiaToolBoxStainNorm(TransformerAdapter):
    """
    Stain normailaztion Adapter, using tiatoolbox.tools.stainnorm
    """
    def __init__(self, path_to_target_image:str , method_name: str):
        """
        TODO

        Arguments:
            path_to_target_image: Absolute pathto target image
            method_name: TODO
        """

        super().__init__()
        target = plt.imread(path_to_target_image)
        target = convert_to_uint8_image(target)
        self.stain_normalizer = stainnorm.get_normaliser(method_name)
        self.stain_normalizer.fit(target)
        if hasattr(self.stain_normalizer, "target_concentrations"):
            self.stain_normalizer.target_concentrations = None
        self.is_callable_with_batches = True

    def __call__(self, img_or_imgs):
        """
        TODO

        Arguments:
            img_or_imgs: TODO        
            
        Returns:
            TODO: TODO
        """
        if isinstance(img_or_imgs, list):
            result = []
            for img in img_or_imgs:
                img = convert_to_uint8_image(img)
                normed_sample = self.stain_normalizer.transform(img.copy())
                result.append(normed_sample)
            return result
        else:
            img_or_imgs = convert_to_uint8_image(img_or_imgs)
            return self.stain_normalizer.transform(img_or_imgs.copy())



import os
from typing import Callable
from typing import Optional as O
from typing import Sequence
from typing import Union as U

from ..base import data_readers, df_creators, df_manipulators
from ..base.histo_dataset import HistoDataset


def bach_patches(
    root: str,  # '/data/ldap/histopathologic/original_read_only/BACH/ICIAR2018_BACH_Challenge'
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    subset: O[str] = "all",
    splitting_seed: int = 69,
    splitting_trainings_percentage: float = 0.8,
    seed: O[int] = None,
    patch_size: O[int] = None,
    overlapping_percentage: float = 0.0,
) -> HistoDataset:
    """
    Creates the Bach dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at TODO

    Arguments:
        root: absolute path to the dataset files.
                transformation (Callable): A callable or a list of
                       callable transformations.
        pre_transformation (Callable): A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        subset: Which part of the dataset should be used.
                One of {"train", "valid", "test", "all"}
        splitting_seed: Seed used for splitting the data set
        splitting_trainings_percentage: Percentage how to split data
        seed:   The seed that is used for the "pre_transformation" and the
                "transformation". If the seed is set, the "pre_transformation"
                will always use the same seed for a data row. If the seed is
                set, the "transformation" will use always the same seed
                depending on the call position.
        patch_size: TODO
        overlapping_percentage: TODO

    Returns:
        HistoDataset:
            The dataset that loads the bach patches. You can use this
            as normal pytorch dataset. If you call it, the data will
            be returned as dictionary.
    """

    # TODO what does this exactly do?
    if patch_size:
        manipulators_to_use = [
            df_manipulators.EditDFImageGrid(patch_size, overlapping_percentage, (1536, 2048))
        ]
        data_reader = data_readers.ReadFromImageFile(
            "Photos/{label}/{file_name}", return_image_size=patch_size
        )
    else:
        manipulators_to_use = []
        data_reader = data_readers.ReadFromImageFile("Photos/{label}/{file_name}")

    # handle training, validation and test cases
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        manipulators_to_use.append(
            df_manipulators.RandomFilterByColumnValue(
                "file_name", splitting_trainings_percentage, mode=0, seed=splitting_seed
            )
        )
    elif subset.startswith("val"):
        manipulators_to_use.extend(
            [
                df_manipulators.RandomFilterByColumnValue(
                    "file_name", splitting_trainings_percentage, mode=1, seed=splitting_seed
                ),
                df_manipulators.RandomFilterByColumnValue(
                    "file_name", 0.5, mode=0, seed=splitting_seed + 10000
                ),
            ]
        )
    elif subset.startswith("te"):
        manipulators_to_use.extend(
            [
                df_manipulators.RandomFilterByColumnValue(
                    "file_name", splitting_trainings_percentage, mode=1, seed=splitting_seed
                ),
                df_manipulators.RandomFilterByColumnValue(
                    "file_name", 0.5, mode=1, seed=splitting_seed + 10000
                ),
            ]
        )
    elif subset == "all":
        pass
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "valid", "test", "all"].'
        )

    # path to ground truth
    path_to_csv = "Photos/microscopy_ground_truth.csv"

    # create HistoDataset object and pass relevant attributes
    ds = HistoDataset(
        df_creators.CreateDFFromCSV(path_to_csv, header=None, names=["file_name", "label"]),
        root,
        df_manipulators=manipulators_to_use,
        data_readers=data_reader,
        feature_readers=data_readers.ReadValueFromCSV(
            r"{label}", encoded_values=["Benign", "InSitu", "Invasive", "Normal"]
        ),
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds


""" TODO: how the offline dataset was generated
from torchvision import transforms

transform = transforms.Compose([
    transforms.ToPILImage(),
    transforms.Resize(423), # (2*299**2)**0.5 => 423
    transforms.ToTensor(),
])

# BACH
from histodata.datasets import bach

path_bach = '/data/ldap/histopathologic/original_read_only/BACH/ICIAR2018_BACH_Challenge'

ds = bach.bach_patches(path_bach,
                       transformation=transform,
                       subset='train',
                       seed=200,
                      )
ds.save('/data/ldap/histopathologic/original_read_only/BACH2/train')
ds = bach.bach_patches(path_bach,
                       transformation=transform,
                       subset='valid',
                       seed=200,
                      )
ds.save('/data/ldap/histopathologic/original_read_only/BACH2/valid')
ds = bach.bach_patches(path_bach,
                       transformation=transform,
                       subset='test',
                       seed=200,
                      )
ds.save('/data/ldap/histopathologic/original_read_only/BACH2/test')
"""


def bach_patches_preprocessed(
    root: str,  # '/data/ldap/histopathologic/original_read_only/BACH2/'
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    subset: O[str] = "train",
    seed: O[int] = None,
) -> HistoDataset:
    """
    Creates the Bach dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at TODO

    Arguments:
        root: absolute path to the dataset files.
                transformation (Callable): A callable or a list of
                       callable transformations.
        pre_transformation (Callable): A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        subset: Which part of the dataset should be used.
                One of {"train", "valid", "test"}
        seed:   The seed that is used for the "pre_transformation" and the
                "transformation". If the seed is set, the "pre_transformation"
                will always use the same seed for a data row. If the seed is
                set, the "transformation" will use always the same seed
                depending on the call position.

    Returns:
        HistoDataset:
            The dataset that loads the bach patches. You can use this
            as normal pytorch dataset. If you call it, the data will
            be returned as dictionary.
    """
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        path_to_dataset = os.path.join(root, "train")
    elif subset.startswith("val"):
        path_to_dataset = os.path.join(root, "valid")
    elif subset.startswith("test"):
        path_to_dataset = os.path.join(root, "test")
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "valid", "test"].'
        )

    # create HistoDataset object and pass relevant attributes
    ds = HistoDataset(
        df_creators.CreateDFFromCSV("features.csv"),
        path_to_dataset,
        data_readers="data/{id}.tiff",
        feature_readers=data_readers.ReadValueFromCSV(
            r"{feature}", encoded_values=["0", "1", "2", "3"]
        ),
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds



import os
from typing import Callable
from typing import Optional as O
from typing import Sequence
from typing import Union as U

from ..base import data_readers, df_creators, df_manipulators
from ..base.histo_dataset import HistoDataset


def mai_patches(
    root: str,
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    folds: O[Sequence[int]] = None,
    subset: O[str] = None,
    gt_mode: str = "GT3&GT2",
    stain_mode: str = "phh3",
    seed: O[int] = None,
) -> HistoDataset:
    """
    The functions returns a pytorch dataset to load the MAI patches. The
    MAI patches was created in the MAI project. Pathologist annotate
    mitosis on five PHH3 WSIs. The corresponding HE slide was registered
    to he PHH3 tile. The patches have the size of 512x512 pixels and the
    label none-mitosis, G2 stadium or mitosis stands for the cell in the
    center of the image.

    Args:
        root (str): The path to the dataset.
        transformation (Callable): A callable or a list of
                       callable transformations.
        pre_transformation (Callable): A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        folds (list(int)): The dataset is splitted into 10 folds. The
                       last (the 9th) should be used only for testing.
                       The list contains numbers between 0 and 9.
        subset: which part of the dataset should be used.
                       One of {"train", "validation", "test", "all"}
                       This parameter is only used if folds is not set.
                       "train" equals folds: [0,1,2,3,4,5,6],
                       "validation" equals folds: [7,8],
                       "test" equals folds: [9]
        gt_mode (str): The pathologist take a look at the slides three times
                       (GT1, GT2, GT3). One of the following modes are allowed:
                       - "GT1": 'None-Mitosis' / 'Mitosis' from the first study
                                only at tumor regions
                       - "GT2": 'None-Mitosis' / 'Mitosis' from the second study
                                only at non-tumor regions
                       - "GT3": 'None-Mitosis' / 'G2' / 'Mitosis' from the
                                third study, only at tumor regions
                       - "GT1&GT2": Use GT1 in tumor regions and G2 in
                                none-tumor regions.
                       - "GT3&GT2": Use GT3 (with G2 as Non-Mitosis)
                                in tumor regions and G2 in non-tumor regions.
                       - "GT1=GT3": 'Non-Mitosis' / 'Mitosis'. The label mitosis
                                is only given if the first and the third study
                                marked the same region as mitosis.
                                Only tumor regions are returned.
                       - "GT1=GT3&GT2": 'Non-Mitosis' / 'Mitosis'. Same as
                                "GT1=GT3" for tumor region + "GT2" for
                                none-tumor regionen
                       - "without": no label will be returned.
        stain_mode (str): The stain to return, one of "phh3", "he", "he&phh3"
        seed (int): The seed that is used for the "pre_transformation" and the
                    "transformation". If the seed is set, the "pre_transformation"
                    will always use the same seed for a data row. If the seed is
                    set, the "transformation" will use always the same seed
                    depending on the call position. For example:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612
                    If we create the exact same dataset the transformation seeds
                    will be the same:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612


    Returns:
        HistoDataset: The dataset that loads the mai patches. You can use this
                      as normal pytorch dataset. If you call it you will get
                      a dictionary back.
    """
    if folds is None and subset is not None:
        subset = subset.lower()
        if subset.startswith("tr"):
            folds = [0, 1, 2, 3, 4, 5, 6]
        elif subset.startswith("val"):
            folds = [7, 8]
        elif subset.startswith("te"):
            folds = [9]
        elif subset == "all":
            pass
        else:
            raise AttributeError(
                "Such type is not supported. Please choose the subset 'train', 'validation' or 'test'."
            )

    if folds is not None:
        manipulators_to_use = [df_manipulators.ColumnFilter("fold", folds)]
    else:
        manipulators_to_use = []

    if gt_mode == "GT1&GT2":
        feature_reader = data_readers.ReadValueFromCSV(r"mitosis", encoded_values=[False, True])
    elif gt_mode == "GT3&GT2":

        def get_gt3_and_gt2(row):
            if row["area"] == "tumor":
                if row["GT3"] == "M":
                    return 1
                else:
                    return 0
            else:
                if row["mitosis"]:
                    return 1
                else:
                    return 0

        feature_reader = data_readers.ReadValueFromCSV(get_gt3_and_gt2)

    elif gt_mode == "GT1=GT3&GT2":

        def get_gt1_eq_gt3_and_gt2(row):
            if row["area"] == "tumor":
                if row["GT1_3"] == "True_M":
                    return 1
                else:
                    return 0
            else:
                if row["mitosis"]:
                    return 1
                else:
                    return 0

        feature_reader = data_readers.ReadValueFromCSV(get_gt1_eq_gt3_and_gt2)
    elif gt_mode == "GT1=GT3":
        feature_reader = data_readers.ReadValueFromCSV(r"GT1_3", encoded_values=["True_M"])
        manipulators_to_use.append(df_manipulators.ColumnFilter("area", ["tumor"]))
    elif gt_mode == "GT1":
        feature_reader = data_readers.ReadValueFromCSV(r"mitosis", encoded_values=[False, True])
        manipulators_to_use.append(df_manipulators.ColumnFilter("area", ["tumor"]))
    elif gt_mode == "GT2":
        feature_reader = data_readers.ReadValueFromCSV(r"mitosis", encoded_values=[False, True])
        manipulators_to_use.append(df_manipulators.ColumnFilter("area", ["other"]))
    elif gt_mode == "GT3":
        feature_reader = data_readers.ReadValueFromCSV(r"GT3", encoded_values=["other", "G2", "M"])
        manipulators_to_use.append(df_manipulators.ColumnFilter("GT3", ["other", "G2", "M"]))
    elif gt_mode == "without":
        feature_reader = None
    else:
        raise NotImplementedError(
            'The gt_mode must be one of the following values: ["GT1", "GT2", '
            + '"GT3", "GT1&GT2", "GT3&GT2", "GT1=GT3", "GT1=GT3&GT2"]. The'
            + ' default value is "GT3&GT2".'
        )

    if stain_mode == "phh3":
        readers_to_use = "PHH3/{id}.png"
    elif stain_mode == "he":
        readers_to_use = "HE/{id}.png"
    elif stain_mode == "he&phh3":
        readers_to_use = ["HE/{id}.png", "PHH3/{id}.png"]
    else:
        raise NotImplementedError(
            "The stain_mode must be one of the following values: "
            + '["phh3", "he", "he&phh3"]. The default value is phh3'
        )

    ds = HistoDataset(
        "data4e.csv",
        root,
        df_manipulators=manipulators_to_use,
        data_readers=readers_to_use,
        feature_readers=feature_reader,
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds


def mai_patches_bern(
    root: str,
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    stain_mode: str = "phh3",
    seed: O[int] = None,
):
    r"""
    The functions returns a pytorch dataset to load the MAI patches
    that was created from bern. There exists no labels to this dataset.

    Args:
        root (str): The path to the dataset.
        transformation (Callable): A callable or a list of
                       callable transformations.
        pre_transformation (Callable): A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       results.
        stain_mode (str): currently only the phh3 stain exists.
        seed (int): The seed that is used for the "pre_transformation" and the
                    "transformation". If the seed is set, the "pre_transformation"
                    will always use the same seed for a data row. If the seed is
                    set, the "transformation" will use always the same seed
                    depending on the call position. For example:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612
                    If we create the exact same dataset the transformation seeds
                    will be the same:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612

    Returns:
        HistoDataset: The dataset that loads the mai patches. You can use this
                      as normal pytorch dataset. If you call it you will get
                      a dictionary back.
    """
    if stain_mode == "phh3":
        readers_to_use = "PHH3/{base}.mrxs_{x}_{y}.png"
    elif stain_mode == "he_just_a_tmp_do_not_use_this":
        readers_to_use = "HE/{base}.mrxs_{x}_{y}.png"
    elif stain_mode == "he&phh3_just_a_tmp_do_not_use_this":
        readers_to_use = ["HE/{base}.mrxs_{x}_{y}.png", "PHH3/{base}.mrxs_{x}_{y}.png"]
    else:
        raise NotImplementedError(
            "The stain_mode must be one of the following values: "
            + '["phh3"]. The default value is phh3'
        )

    ds = HistoDataset(
        df_creators.CreateDFFromFolder("PHH3/{base}.mrxs_{x}_{y}.png"),
        root,
        df_manipulators=None,
        data_readers=readers_to_use,
        feature_readers=None,
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds


def mai_he_and_phh3_fields(
    root: str,
    subset: str,
    splitting_seed: O[int] = 69,
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    patch_size: O[int] = 224,
    overlapping_percentage: O[float] = 0.0,
    seed=None,
    stain_mode="he&phh3",
    splitting_trainings_percentage=0.8,
):
    r"""
    The functions returns a pytorch dataset to load the MAI patches. The
    MAI patches was created in the MAI project. The fields are 2048x2048
    and the stain HE and PHH3 are registered to each other. The dataset
    that is here created grids the fields into the given patch size.

    Args:
        root (str): The path to the dataset.
        subset (str): which part of the dataset should be used.
                       One of {"train", "validation", "test", "all"}
        splitting_seed (int): The splitting seed that is used to create
                       the datasets.
        splitting_trainings_percentage (float): 0.0 < x < 1.0, the percentage
                       that is used for the split the train part from the
                       validation/test part. The validation and test is
                       equally splitted.
        transformation (Callable): A callable or a list of
                       callable transformations.
        pre_transformation (Callable): A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        patch_size (int): The size that is used for the grid. The image get
                       extracted before the transformation is called. If None,
                       the original size is returned.
        overlapping_percentage (float): 0.0 <= x < 1.0. The percentage that
                       is used for the overlapping. The value 0.5 means that
                       50 percent of one patch to the next patch are the same.
        stain_mode (str): The stain to return, one of "phh3", "he", "he&phh3"
        seed (int): The seed that is used for the "pre_transformation" and the
                    "transformation". If the seed is set, the "pre_transformation"
                    will always use the same seed for a data row. If the seed is
                    set, the "transformation" will use always the same seed
                    depending on the call position. For example:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612
                    If we create the exact same dataset the transformation seeds
                    will be the same:
                    Call-Index,     Seed f.              Seed f.
                                pre_transformation,  transformation
                        0     ,     100           ,    124
                        1     ,     100           ,    512
                        2     ,     100           ,    810
                        3     ,     100           ,    612

    Returns:
        HistoDataset: The dataset that loads the mai patches. You can use this
                      as normal pytorch dataset. If you call it you will get
                      a dictionary back.
    """
    if patch_size is None:
        manipulators_to_use = []
    else:
        manipulators_to_use = [
            df_manipulators.EditDFImageGrid(patch_size, overlapping_percentage, (2048, 2048))
        ]
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        manipulators_to_use.append(
            df_manipulators.RandomFilterByColumnValue(
                "base", splitting_trainings_percentage, mode=0, seed=splitting_seed
            )
        )
    elif subset.startswith("val"):
        manipulators_to_use.extend(
            [
                df_manipulators.RandomFilterByColumnValue(
                    "base", splitting_trainings_percentage, mode=1, seed=splitting_seed
                ),
                df_manipulators.RandomFilterByColumnValue(
                    "base", 0.5, mode=0, seed=splitting_seed + 10000
                ),
            ]
        )
    elif subset.startswith("te"):
        manipulators_to_use.extend(
            [
                df_manipulators.RandomFilterByColumnValue(
                    "base", splitting_trainings_percentage, mode=1, seed=splitting_seed
                ),
                df_manipulators.RandomFilterByColumnValue(
                    "base", 0.5, mode=1, seed=splitting_seed + 10000
                ),
            ]
        )
    elif subset == "all":
        pass
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "validation", "test", "all"].'
        )

    if stain_mode == "phh3":
        readers_to_use = data_readers.ReadFromImageFile(
            "{base}/{base}.phh3.png", return_image_size=patch_size
        )
    elif stain_mode == "he":
        readers_to_use = data_readers.ReadFromImageFile(
            "{base}/{base}.he.png", return_image_size=patch_size
        )
    elif stain_mode == "he&phh3":
        readers_to_use = [
            data_readers.ReadFromImageFile("{base}/{base}.he.png", return_image_size=patch_size),
            data_readers.ReadFromImageFile("{base}/{base}.phh3.png", return_image_size=patch_size),
        ]
    else:
        raise NotImplementedError(
            "The stain_mode must be one of the following values: "
            + '["phh3", "he", "he&phh3"]. The default value is phh3'
        )

    ds = HistoDataset(
        df_creators.CreateDFFromFolder("{base}/{base}.he.png"),
        root,
        data_readers=readers_to_use,
        df_manipulators=manipulators_to_use,
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds


def mai_slides(
    root,  # '/data/ldap/histopathologic/original_read_only/MAI/slides
    transformation=None,
    pre_transformation=None,
    patch_size=224,
    overlapping_percentage=0.0,  # TODO
    seed=None,
    stain_mode="all",  # one of ["he", "phh3", "all"]
    scanner_mode="all",  # one of ["xr", "p1000", "p150", "all"]
    location_mode="all",  # one of ["bern", "charite", "all"]
    mpp=0.5,
):
    if stain_mode == "all":
        stain = "{stain}"

    elif stain_mode in ("he", "phh3"):
        stain = stain_mode
    else:
        raise NotImplementedError()

    if location_mode == "all":
        location = "{location}"
    elif location_mode in ("bern", "charite"):
        location = stain_mode
    else:
        raise NotImplementedError()

    if scanner_mode == "all":
        scanner = "{scanner}"
    elif scanner_mode in ("xr", "p1000", "p150"):
        scanner = scanner_mode
    else:
        raise NotImplementedError()

    ds = HistoDataset(
        df_creators.CreateDFFromFolder(
            location + "/" + stain + "/" + scanner + "/{name}\\.(ndpi|mrxs)"
        ),
        root,
        df_manipulators=df_manipulators.EditDFWSIGrid(
            patch_size, perc_overlapping=overlapping_percentage
        ),
        data_readers=data_readers.ReadFromWSI(patch_size, "__path__", mpp=mpp),
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds



import os
from typing import Callable
from typing import Optional as O
from typing import Sequence
from typing import Union as U

import numpy as np

from ..base import data_readers, df_creators, df_manipulators
from ..base.histo_dataset import HistoDataset


def midog_classification_patches(
    root: str,  # '/data/ldap/histopathologic/original_read_only/MIDOG'
    transformation=None,
    pre_transformation=None,
    subset: str = "all",
    splitting_seed: int = 69,
    splitting_trainings_percentage: float = 0.8,
    scanner_to_use: U[int, list] = None,
    seed: int = None,
    patch_size: int = None,
) -> HistoDataset:
    """
    Creates the MIDOG dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at TODO.

    Arguments:
        root: Absolute path to the dataset files.
        transformation: A callable or a list of callable transformations.
        pre_transformation: A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        subset: Which part of the dataset should be used.
                One of {"train", "validation", "test", "all"}
        splitting_seed: Seed used for splitting the data set
        splitting_trainings_percentage: Percentage how to split data
        scanner_to_use: TODO
        seed:   The seed that is used for the "pre_transformation" and the
                "transformation". If the seed is set, the "pre_transformation"
                will always use the same seed for a data row. If the seed is
                set, the "transformation" will use always the same seed
                depending on the call position.
        patch_size: TODO

    Returns:
        HistoDataset:
            The dataset that loads the bach patches. You can use this
            as normal pytorch dataset. If you call it, the data will
            be returned as dictionary.
    """
    manipulators_to_use = []

    # handle training, validation and test cases
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        manipulators_to_use.append(
            df_manipulators.RandomFilterByColumnValue(
                "image_id", splitting_trainings_percentage, mode=0, seed=splitting_seed
            )
        )
    elif subset.startswith("val"):
        manipulators_to_use.extend(
            [
                df_manipulators.RandomFilterByColumnValue(
                    "image_id", splitting_trainings_percentage, mode=1, seed=splitting_seed
                ),
                df_manipulators.RandomFilterByColumnValue(
                    "image_id", 0.5, mode=0, seed=splitting_seed + 10000
                ),
            ]
        )
    elif subset.startswith("te"):
        manipulators_to_use.extend(
            [
                df_manipulators.RandomFilterByColumnValue(
                    "image_id", splitting_trainings_percentage, mode=1, seed=splitting_seed
                ),
                df_manipulators.RandomFilterByColumnValue(
                    "image_id", 0.5, mode=1, seed=splitting_seed + 10000
                ),
            ]
        )
    elif subset == "all":
        pass
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "validation", "test", "all"].'
        )

    # handle scanner values
    if scanner_to_use:
        if isinstance(scanner_to_use, int):
            allowed_values = np.arange(50) + 1 + 50 * scanner_to_use
        elif isinstance(scanner_to_use, list):
            allowed_values = []
            for s in scanner_to_use:
                allowed_values.extend(np.arange(50) + 1 + 50 * s)
        else:
            raise NotImplementedError(
                'The parameter "scanner_to_use" needs to be None, an integer or a list.'
            )
        manipulators_to_use.append(df_manipulators.ColumnFilter("image_id", allowed_values))

    manipulators_to_use.append(df_manipulators.BoundingBoxToCenterCoordinates("bbox"))

    data_reader = data_readers.ReadFromImageFile(
        "images/{image_id:03d}.tiff", return_image_size=patch_size
    )

    # create HistoDataset object and pass relevant attributes
    ds = HistoDataset(
        df_creators.CreateDFFromJSON("MIDOG.json", "annotations"),
        root,
        df_manipulators=manipulators_to_use,
        data_readers=data_reader,
        feature_readers=data_readers.ReadValueFromCSV(r"{category_id}", encoded_values=["1", "2"]),
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds


""" TODO: how the offline dataset was generated
# MIDOG
from histodata.datasets import midog

from torchvision import transforms

transform = transforms.Compose([
    transforms.ToPILImage(),
    #transforms.Resize(423), # (2*299**2)**0.5 => 423
    transforms.ToTensor(),
])

path_midog = '/data/ldap/histopathologic/original_read_only/MIDOG'

ds = midog.midog_classification_patches(path_midog,
                       transformation=transform,
                       subset='train',
                       seed=200,
                       patch_size=423, # (2*299**2)**0.5 => 423,
                      )
ds.save('/data/ldap/histopathologic/original_read_only/MIDOG2/train')

ds = midog.midog_classification_patches(path_midog,
                       transformation=transform,
                       subset='valid',
                       seed=200,
                       patch_size=423, # (2*299**2)**0.5 => 423,
                      )
ds.save('/data/ldap/histopathologic/original_read_only/MIDOG2/valid')

ds = midog.midog_classification_patches(path_midog,
                       transformation=transform,
                       subset='test',
                       seed=200,
                       patch_size=423, # (2*299**2)**0.5 => 423,
                      )
ds.save('/data/ldap/histopathologic/original_read_only/MIDOG2/test')
"""


def midog_classification_patches_preprocessed(
    root: str,  # '/data/ldap/histopathologic/original_read_only/MIDOG2/'
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    subset: O[str] = "train",
    seed: O[int] = None,
) -> HistoDataset:
    """
    Creates the midog dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at TODO

    Arguments:
        root: absolute path to the dataset files.
                transformation (Callable): A callable or a list of
                       callable transformations.
        pre_transformation (Callable): A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        subset: Which part of the dataset should be used.
                One of {"train", "valid", "test"}
        seed:   The seed that is used for the "pre_transformation" and the
                "transformation". If the seed is set, the "pre_transformation"
                will always use the same seed for a data row. If the seed is
                set, the "transformation" will use always the same seed
                depending on the call position.

    Returns:
        HistoDataset:
            The dataset that loads the bach patches. You can use this
            as normal pytorch dataset. If you call it, the data will
            be returned as dictionary.
    """
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        path_to_dataset = os.path.join(root, "train")
    elif subset.startswith("val"):
        path_to_dataset = os.path.join(root, "valid")
    elif subset.startswith("test"):
        path_to_dataset = os.path.join(root, "test")
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "valid", "test"].'
        )

    # create HistoDataset object and pass relevant attributes
    ds = HistoDataset(
        df_creators.CreateDFFromCSV("features.csv"),
        path_to_dataset,
        data_readers="data/{id}.tiff",
        feature_readers=data_readers.ReadValueFromCSV(r"{feature}", encoded_values=["0", "1"]),
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds



from typing import Dict
from typing import Optional as O
from typing import Sequence
from typing import Union as U

import sklearn
import sklearn.metrics
import torch


def pretty_print_confusion_matrix(cm):
    print(" Confusion Matrix ".center(26, "="))
    print("       " + "Actual Value".center(8 * len(cm), " "))
    for idx, row in enumerate(cm):
        if idx == 0:
            print("Pred.  ", end="")
        elif idx == 1:
            print("Value  ", end="")
        else:
            print("       ", end="")
        for _, element in enumerate(row):
            print(str(element).center(8, " "), end="")
        print()


class MetricGenerator:
    def __init__(
        self,
        num_of_labels: int = 2,
        *,
        return_accuracy: bool = True,
        return_details_for_classes: O[U[int, Sequence[int]]] = None,
        print_confusion_matrix: bool = False,
        return_recall: bool = False,
        return_precission: bool = False,
        return_f1: bool = False,
        return_roc: bool = False,
        # return_frocs=None,  # NotImplemented
        pre_name: O[str] = None,
    ) -> None:

        self.num_of_labels = num_of_labels
        self.return_accuracy = return_accuracy
        if isinstance(return_details_for_classes, list):
            self.return_details_for_classes = return_details_for_classes
        else:
            self.return_details_for_classes = []
        self.print_confusion_matrix = print_confusion_matrix
        self.return_recall = return_recall
        self.return_precission = return_precission
        self.return_f1 = return_f1
        self.return_roc = return_roc

        # if isinstance(return_frocs, list):
        #     self.return_frocs = return_frocs
        # else:
        #     self.return_frocs = []

        if pre_name == "":
            self.pre_name = None
        else:
            self.pre_name = pre_name

    def create_metric(self, predictions: torch.Tensor, labels: torch.Tensor) -> Dict[str, float]:
        pred = predictions.argmax(dim=1, keepdim=True)
        pred_eq = pred.eq(labels.view_as(pred))

        # the metric to return
        m = {}

        m["cross_entropy_loss"] = torch.nn.functional.cross_entropy(predictions, labels).item()

        if self.return_accuracy:
            m["accuracy"] = pred_eq.sum().item() / len(labels)

            for lbl in range(self.num_of_labels):
                if len(labels[labels == lbl]) > 0:
                    m[f"accuracy_c{lbl}"] = pred_eq[labels == lbl].sum().item() / len(
                        labels[labels == lbl]
                    )

        if self.print_confusion_matrix:
            cm = sklearn.metrics.confusion_matrix(pred, labels)
            pretty_print_confusion_matrix(cm)

        for lbl in self.return_details_for_classes:
            tp = torch.sum(pred_eq[labels == lbl]).item()
            fn = torch.sum(~pred_eq[labels == lbl]).item()
            # tn = torch.sum(pred[labels != lbl] != lbl).item()
            fp = torch.sum(pred[labels != lbl] == lbl).item()

            precision = None
            if (tp + fp) > 0:
                precision = tp / (tp + fp)
                if self.return_precission:
                    m[f"precision_c{lbl}"] = precision

            recall = None
            if (tp + fn) > 0:
                recall = tp / (tp + fn)
                if self.return_recall:
                    m[f"recall_c{lbl}"] = recall

            if self.return_f1 and precision is not None and recall is not None and m[f"precision_c{lbl}"] + m[f"recall_c{lbl}"] > 0.0:
                m[f"F1_c{lbl}"] = (2.0 * m[f"precision_c{lbl}"] * m[f"recall_c{lbl}"]) / (
                    m[f"precision_c{lbl}"] + m[f"recall_c{lbl}"]
                )

            if self.return_roc:
                # calculate the fpr and tpr for all thresholds of the classification
                fpr, tpr, _ = sklearn.metrics.roc_curve(labels == lbl, predictions[:, lbl])
                m[f"roc_auc_c{lbl}"] = sklearn.metrics.auc(fpr, tpr)

        if self.pre_name is not None:
            return {self.pre_name + e: m[e] for e in m}
        else:
            return m



import os
import random

import imagesize
import numpy as np
import pandas as pd
from PIL import Image
from tiatoolbox.dataloader import wsireader

from ..helper import helper


class Reader:
    """
    Parent class for all readers. Subclasses need to implement their call method.
    """

    def __init__(self):
        pass

    def __call__(self, row, path_to_image):
        raise NotImplementedError


class ReadTestImageV2(Reader):
    """
    TODO
    """

    def __init__(
        self,
        image_size: int = 100,
        to_hash_added: int = 0,
        num_of_circles: int = 10,
        return_mask: bool = False,
        return_coordinats: bool = False,
        return_bbox: bool = False,
    ):
        """
        TODO

        Arguments:
            image_size: TODO
            to_hash_added: TODO
            num_of_circles: TODO
            return_mask: TODO
            return_coordinats: TODO
            return_bbox: TODO
        """

        super().__init__()
        self.img_size = image_size
        self.to_hash_added = to_hash_added
        self.return_mask = return_mask
        self.return_coordinats = return_coordinats
        self.return_bbox = return_bbox
        self.num_of_circles = num_of_circles

    def create_test_image(self, row: int) -> dict:
        """
        TODO

        Arguments:
            row: TODO

        Returns:
            TODO: TODO
        """

        img = np.ones([self.img_size, self.img_size, 3], dtype=np.uint8) * 255

        hash_of_row = row["__hash_of_row__"] + self.to_hash_added

        r = random.Random(hash_of_row)

        result = {}

        for i in range(self.num_of_circles):
            coordinates = (r.randint(5, self.img_size - 5), r.randint(5, self.img_size - 5))
            if "coordinates" in result:
                result["coordinates"].append(coordinates)
            else:
                result["coordinates"] = [coordinates]
            radius = r.randint(10, 25)
            mask = helper.create_circular_mask(self.img_size, self.img_size, coordinates, radius)

            img[mask] -= 128
            img[mask, r.randint(0, 2)] -= 128

            if "mask" in result:
                result["mask"] += np.array(mask, dtype=np.uint8)
            else:
                result["mask"] = np.array(mask, dtype=np.uint8)
            bbox = [
                coordinates[0] - radius,
                coordinates[1] - radius,
                coordinates[0] + radius,
                coordinates[1] + radius,
                i,
            ]
            if bbox[0] < 0:
                bbox[0] = 0
            if bbox[1] < 0:
                bbox[1] = 0
            if bbox[2] >= self.img_size:
                bbox[2] = self.img_size - 1
            if bbox[3] >= self.img_size:
                bbox[3] = self.img_size - 1
            if "bbox" in result:
                result["bbox"].append(bbox)
            else:
                result["bbox"] = [bbox]
        result["img"] = img
        if self.return_mask is False:
            result.pop("mask")
        if self.return_coordinats is False:
            result.pop("coordinates")
        if self.return_bbox is False:
            result.pop("bbox")
        return result

    def __call__(self, row_or_df, path_to_image: str):
        """
        TODO

        Arguments:
            row_or_df: TODO
            path_to_image: Absolute path to to image.

        Returns:
            TODO: TODO
        """

        if isinstance(row_or_df, pd.DataFrame):
            imgs = {}

            for _, row in row_or_df.iterrows():
                result = self.create_test_image(row)
                for k in result:
                    if k in imgs:
                        imgs[k].append(result[k])
                    else:
                        imgs[k] = [result[k]]
            return imgs
        else:
            return self.create_test_image(row_or_df)


class ReadTestImage(Reader):
    """
    TODO
    """

    def __init__(
        self,
        image_size: int = 100,
        to_hash_added: int = 0,
        return_mask: bool = False,
        return_coordinats: bool = False,
        return_bbox: bool = False,
    ):
        """
        TODO

        Arguments:
            image_size: TODO
            to_hash_added: TODO
            return_mask: TODO
            return_coordinats: TODO
            return_bbox: TODO

        """

        super().__init__()
        self.img_size = image_size
        self.to_hash_added = to_hash_added
        self.return_mask = return_mask
        self.return_coordinats = return_coordinats
        self.return_bbox = return_bbox

    def create_test_image(
        self,
        row,
    ):
        """
        TODO

        Arguments:
            row: TODO

        Returns:
            TODO: TODO
        """

        img = np.ones([self.img_size, self.img_size, 3], dtype=np.uint8) * 255

        hash_of_row = row["__hash_of_row__"] + self.to_hash_added

        pos = 0
        set_color = 255

        r = random.Random(hash_of_row)

        while pos < self.img_size:
            next_length = r.randint(8, 50)
            pos2 = pos + next_length

            if set_color == 255:
                img[pos:pos2, : int(self.img_size / 2), 0] = set_color
                img[pos:pos2, : int(self.img_size / 2), 1] = r.randint(0, 64)
                img[pos:pos2, : int(self.img_size / 2), 2] = r.randint(0, 64)
            else:
                img[pos:pos2, : int(self.img_size / 2), 0] = r.randint(0, 255)
                img[pos:pos2, : int(self.img_size / 2), 1] = r.randint(0, 64)
                img[pos:pos2, : int(self.img_size / 2), 2] = r.randint(0, 64)

            if set_color == 255:
                img[pos:pos2, int(self.img_size / 2) :, 0] = r.randint(0, 64)
                img[pos:pos2, int(self.img_size / 2) :, 1] = set_color
                img[pos:pos2, int(self.img_size / 2) :, 2] = r.randint(0, 64)
            else:
                img[pos:pos2, int(self.img_size / 2) :, 0] = r.randint(0, 64)
                img[pos:pos2, int(self.img_size / 2) :, 1] = r.randint(0, 255)
                img[pos:pos2, int(self.img_size / 2) :, 2] = r.randint(0, 64)

            pos += next_length
            if set_color == 0:
                set_color = 255
            else:
                set_color = 0

        result = {}
        result["img"] = img
        if self.return_mask:
            result["mask"] = np.array(np.mean(img, axis=2) > 96, dtype=np.uint8)
            result["multi_label_mask"] = (
                1 * (img[:, :, 0] > 128) + 2 * (img[:, :, 1] > 128) + 4 * (img[:, :, 2] > 32)
            )
        if self.return_coordinats:
            result["coordinates"] = []
            for _ in range(10):
                result["coordinates"].append(
                    [r.randint(5, self.img_size - 5), r.randint(5, self.img_size - 5)]
                )
        if self.return_bbox:
            result["bbox"] = []
            for _ in range(3):
                x = r.randint(5, self.img_size - 150)
                y = r.randint(5, self.img_size - 150)
                result["bbox"].append([x, y, x + r.randint(50, 150), y + r.randint(50, 150)])

        return result

    def __call__(self, row_or_df, path_to_image):
        """
        TODO

        Arguments:
            row_or_df: TODO
            path_to_image: TODO

        Returns:
            TODO: TODO
        """

        if isinstance(row_or_df, pd.DataFrame):
            imgs = {}

            for _, row in row_or_df.iterrows():
                result = self.create_test_image(row)
                for k in result:
                    if k in imgs:
                        imgs[k].append(result[k])
                    else:
                        imgs[k] = [result[k]]
            return imgs
        else:
            return self.create_test_image(row_or_df)


def expression_to_string_value(expression, interpreter, row):
    """
    TODO

    Arguments:
        expression: TODO
        interpreter: TODO
        row: TODO

    Returns:
        TODO: TODO
    """

    if expression in row:
        return row[expression]
    elif callable(expression):
        return expression(row)
    elif callable(interpreter):
        return interpreter(row, expression)
    else:
        return expression


class ReadFromImageFile(Reader):
    """
    TODO
    """

    def __init__(
        self,
        expression,
        interpreter=helper.read_from_row,
        return_image_size=None,
        x_column: str = "__x__",
        y_column: str = "__y__",
    ):
        """
        TODO

        Arguments:
            expression: TODO
            interpreter: TODO
            return_image_size: TODO
            x_column: TODO
            y_column: TODO
        """

        super().__init__()
        self.expression = expression
        self.interpreter = interpreter
        self.return_image_size = return_image_size
        self.x_column = x_column
        self.y_column = y_column

    def crop(self, img, x, y):
        """
        TODO

        Arguments:
            img: TODO
            x: TODO
            y: TODO

        Returns:
            TODO: TODO
        """

        shifted_x, shifted_y = 0, 0
        start_x = int(x - self.return_image_size / 2)
        if start_x < 0:
            shifted_x = -start_x
            start_x = 0
        start_y = int(y - self.return_image_size / 2)
        if start_y < 0:
            shifted_y = -start_y
            start_y = 0
        end_x = start_x + self.return_image_size - shifted_x
        end_y = start_y + self.return_image_size - shifted_y
        img = img[start_x:end_x, start_y:end_y]
        if img.shape[0] != self.return_image_size or img.shape[1] != self.return_image_size:
            img_black_boarder = np.zeros(
                [self.return_image_size, self.return_image_size, img.shape[2]], dtype=img.dtype
            )
            img_black_boarder[
                shifted_x : shifted_x + img.shape[0], shifted_y : shifted_y + img.shape[1]
            ] = img
            return img_black_boarder
        else:
            return img

    def load_from_row(self, row, path_to_image):
        """
        TODO

        Arguments:
            row: TODO
            path_to_image: TODO

        Returns:
            TODO: TODO
        """

        path = expression_to_string_value(self.expression, self.interpreter, row)
        path_to_file = os.path.join(path_to_image, path)
        # img = plt.imread(path_to_file)
        img = np.array(Image.open(path_to_file))
        if len(img.shape) == 3 and img.shape[-1] == 4:
            img = img[..., :3]
        if self.return_image_size:
            if self.x_column in row:
                x = row[self.x_column]
            else:
                x = img.shape[0] / 2.0
            if self.y_column in row:
                y = row[self.y_column]
            else:
                y = img.shape[1] / 2.0

            img = self.crop(img, x, y)

        # img = np.swapaxes(img, 0, 2)
        # img = torch.as_tensor(img)
        return img

    def get_image_size(self, row, path_to_image):
        """
        TODO

        Arguments:
            row: TODO
            path_to_image: TODO

        Returns:
            TODO: TODO
        """

        path = expression_to_string_value(self.expression, self.interpreter, row)
        path_to_file = os.path.join(path_to_image, path)
        return imagesize.get(path_to_file)

    def __call__(self, row_or_df, path_to_image):
        """
        TODO

        Arguments:
            row_or_df: TODO
            path_to_image: TODO

        Returns:
            TODO: TODO
        """

        if isinstance(row_or_df, pd.DataFrame):
            imgs = []
            for _, row in row_or_df.iterrows():
                imgs.append(self.load_from_row(row, path_to_image))
            return imgs  # torch.stack(imgs)
        else:
            return self.load_from_row(row_or_df, path_to_image)


class ReadFromWSI(Reader):
    """
    TODO
    """

    def __init__(
        self,
        return_image_size,
        expression,
        interpreter=helper.read_from_row,
        mpp: float = 0.5,
        x_column: str = "__x__",
        y_column: str = "__y__",
        use_center_coordinates: bool = True,
    ):
        """
        TODO

        Arguments:
            return_image_size: TODO
            expression: TODO
            interpreter: TODO
            mpp: TODO
            x_column: TODO
            y_column: TODO
            use_center_coordinates: TODO

        """

        super().__init__()
        if isinstance(return_image_size, int):
            self.return_image_size = (return_image_size, return_image_size)
        else:
            self.return_image_size = return_image_size
        self.expression = expression
        self.interpreter = interpreter
        self.mpp = mpp
        self.x_column = x_column
        self.y_column = y_column
        self.use_center_coordinates = use_center_coordinates

    def load_from_row(self, row, path_to_image):
        """
        TODO

        Arguments:
            row: TODO
            path_to_image: Absolute path to image.

        Returns:
            TODO: TODO
        """

        path = expression_to_string_value(self.expression, self.interpreter, row)
        path_to_file = os.path.join(path_to_image, path)

        wsi_reader = wsireader.get_wsireader(input_img=path_to_file)

        if self.use_center_coordinates is False:
            patch = wsi_reader.read_rect(
                location=(int(row[self.x_column]), int(row[self.y_column])),
                resolution=self.mpp,
                units="mpp",
                size=self.return_image_size,
            )
        else:
            x = row[self.x_column]
            y = row[self.y_column]

            x -= (self.return_image_size[0] / 2) * self.mpp / wsi_reader.info.mpp[0]
            y -= (self.return_image_size[1] / 2) * self.mpp / wsi_reader.info.mpp[0]

            patch = wsi_reader.read_rect(
                location=(int(x), int(y)),
                resolution=self.mpp,
                units="mpp",
                size=self.return_image_size,
            )
        return patch

    def read_thumbnail(self, row, path_to_image, mpp, return_min_mpp=False):
        """
        TODO

        Arguments:
            row: TODO
            path_to_image: TODO
            mpp: TODO
            return_min_mpp: TODO

        Returns:
            TODO: TODO
        """

        path = expression_to_string_value(self.expression, self.interpreter, row)
        path_to_file = os.path.join(path_to_image, path)

        print(path_to_file)

        wsi_reader = wsireader.get_wsireader(input_img=path_to_file)

        wsi_thumb = wsi_reader.slide_thumbnail(resolution=mpp, units="mpp")

        if return_min_mpp:
            return wsi_thumb, wsi_reader.info.mpp
        else:
            return wsi_thumb

    def __call__(self, row_or_df, path_to_image):
        """
        TODO

        Arguments:
            row_or_df: TODO
            path_to_image: TODO

        Returns:
            TODO: TODO
        """

        if isinstance(row_or_df, pd.DataFrame):
            imgs = []
            for _, row in row_or_df.iterrows():
                imgs.append(self.load_from_row(row, path_to_image))
            return imgs  # torch.stack(imgs)
        else:
            return self.load_from_row(row_or_df, path_to_image)


class ReadValueFromCSV(Reader):
    """
    TODO
    """

    def __init__(self, expression, interpreter=helper.read_from_row, encoded_values=None):
        """
        TODO

        Arguments:
            expression: TODO
            interpreter: TODO
            encoded_values: TODO

        """

        super().__init__()
        self.expression = expression
        self.interpreter = interpreter
        self.encoded_values = encoded_values

    def load_from_row(self, row):
        """
        TODO

        Arguments:
            row: TODO

        Returns:
            TODO: TODO
        """

        loaded_value = expression_to_string_value(self.expression, self.interpreter, row)
        if self.encoded_values:
            if loaded_value in self.encoded_values:
                return self.encoded_values.index(loaded_value)
            else:
                return len(self.encoded_values)
        return loaded_value

    def __call__(self, row_or_df, path_to_image):
        """
        TODO

        Arguments:
            row_or_df: TODO
            path_to_image: TODO

        Returns:
            TODO: TODO
        """

        if isinstance(row_or_df, pd.DataFrame):
            values = []
            for _, row in row_or_df.iterrows():
                values.append(self.load_from_row(row))
            return values
        else:
            return self.load_from_row(row_or_df)



import json
import os
import re
from typing import Optional as O

import pandas as pd


class DataFrameCreator:
    """
    Parent class for all dataframes. Subclasses need to implement their call method.
    """

    def __init__(self):
        pass

    def __call__(self, path_to_dataset: str):
        raise NotImplementedError


class CreateDFDummy(DataFrameCreator):
    """
    Creates a dummy DataFrame.
    """

    def __init__(self, num_of_entries: int = 100):
        """
        TODO

        Arguments:
            num_of_entries: TODO
        """

        super().__init__()
        self.num_of_entries = num_of_entries

    def __call__(self, path_to_dataset: str):
        """
        TODO

        Arguments:
            path_to_dataset: TODO
        """
        df = pd.DataFrame({"id": range(self.num_of_entries)})
        df["fold"] = df.apply(lambda row: row["id"] % 10, 1)
        df["is_even"] = df.apply(lambda row: (row["id"] % 2) == 0, 1)
        return df


class CreateDFFromJSON(DataFrameCreator):
    """
    Creates a DataFrame from a JSON file.
    """

    def __init__(self, path_to_json: str, key_to_use: O[str] = None):
        """
        TODO

        Arguments:
            path_to_json: TODO
            key_to_use: TODO
        """

        super().__init__()
        self.path_to_json = path_to_json
        self.key_to_use = key_to_use

    def __call__(self, path_to_dataset: str):
        """
        TODO

        Arguments:
            path_to_dataset: TODO

        Returns:
            TODO: TODO
        """

        if isinstance(self.path_to_json, list):
            df = pd.DataFrame()
            for path_to_json in self.path_to_json:
                path = os.path.join(path_to_dataset, path_to_json)
                with open(path) as f:
                    json_dict = json.load(f)
                if self.key_to_use:
                    tmp_df = pd.DataFrame(json_dict[self.key_to_use])
                else:
                    tmp_df = pd.DataFrame(json_dict)
                tmp_df["__from_csv__"] = path_to_json
                df = df.append(tmp_df)
            return df
        else:
            path = os.path.join(path_to_dataset, self.path_to_json)
            with open(path) as f:
                json_dict = json.load(f)
            if self.key_to_use:
                return pd.DataFrame(json_dict[self.key_to_use])
            else:
                return pd.DataFrame(json_dict)


class CreateDFFromCSV(DataFrameCreator):
    """
    Creates a DataFrame from CSV.
    """

    def __init__(self, path_to_csv: str, sep: str = ",", **argv):
        """
        TODO

        Arguments:
            path_to_csv: TODO
            sep: TODO
            **argv: TODO
        """
        super().__init__()
        self.path_to_csv = path_to_csv
        self.sep = sep
        self.argv = argv

    def __call__(self, path_to_dataset: str):
        """
        TODO

        Arguments:
            path_to_dataset: Absolute path to dataset

        Returns:
            TODO: TODO
        """
        if isinstance(self.path_to_csv, list):
            df = pd.DataFrame()
            for path_to_csv in self.path_to_csv:
                path = os.path.join(path_to_dataset, path_to_csv)
                tmp_df = pd.read_csv(path, sep=self.sep, **self.argv)
                tmp_df["__from_csv__"] = path_to_csv
                df = df.append(tmp_df)
            return df
        else:
            path = os.path.join(path_to_dataset, self.path_to_csv)
            return pd.read_csv(path, sep=self.sep, **self.argv)


class CreateDFFromFolder(DataFrameCreator):
    """
    Creates DataFrame based on a folder content.
    """

    def __init__(self, file_of_interest: str):
        """
        TODO

        Arguments:
            file_of_interest: TODO
        """

        super().__init__()
        self.file_of_interest = file_of_interest

        regex = ""
        tmp = file_of_interest
        self.var_names = []
        counter_unnamed = 0
        while len(tmp) > 0:
            var_start = tmp.find("{")
            var_unnamed_start = tmp.find("(")
            if (
                var_unnamed_start != -1
                and (var_unnamed_start < var_start or var_start == -1)
                and (var_unnamed_start == 0 or tmp[var_unnamed_start - 1] != "\\")
            ):
                regex += tmp[: var_unnamed_start + 1]
                tmp = tmp[var_unnamed_start + 1 :]
                self.var_names.append("unnamed_" + str(counter_unnamed))
                counter_unnamed += 1
            elif var_start == -1:  # No Variable could be found
                regex += tmp
                tmp = ""
            else:
                # ignore this var start
                if var_start > 0 and tmp[var_start - 1] == "\\":
                    regex += tmp[:var_start]
                    tmp = tmp[var_start:]
                else:
                    regex += tmp[:var_start]
                    tmp = tmp[var_start:]
                    var_end = tmp.find("}")
                    self.var_names.append(tmp[1:var_end])
                    # self.regex += '(?=.*)'
                    # regex += '(^[^/]*)'
                    regex += "(.*)"
                    tmp = tmp[var_end + 1 :]
        self.regex = re.compile(regex)

    def __call__(self, path_to_dataset: str):
        """
        TODO

        Arguments:
            path_to_dataset: Absolute path to dataset

        Returns:
            TODO: TODO
        """

        rows = {name: [] for name in self.var_names}
        rows["__path__"] = []
        for root, _, files in os.walk(path_to_dataset):
            for file in files:
                path = os.path.join(root, file)[len(path_to_dataset) :]
                if path.startswith("/"):
                    path = path[1:]
                result = self.regex.search(path)
                if result:
                    path_as_variable = False
                    row = {}
                    for vname, var in zip(self.var_names, result.groups()):
                        row[vname] = var
                        if var and var.find("/") != -1:
                            path_as_variable = True
                    if path_as_variable is False:
                        for k in row:
                            rows[k].append(row[k])
                        rows["__path__"].append(path)

        return pd.DataFrame(rows)



import numpy as np
import torch
import torch.utils


class SamplerSimple(torch.utils.data.Sampler[int]):
    r"""Samples elements sequentielly.

    Arguments:
        num_of_samples: The number of data samples that the dataset have.
        batch_size: the number of ids that this sampler should return in one call.
        shuffle: If true, the elements get shuffled each epoch.
    """

    def __init__(self, num_of_samples: int, batch_size: int = 64, shuffle: bool = False) -> None:
        super().__init__(data_source=None)

        self.num_of_samples = num_of_samples
        self.batch_size = batch_size
        self.shuffle = shuffle

    @property
    def num_samples(self) -> int:
        return self.num_of_samples

    def __iter__(self):
        indices = np.arange(self.num_of_samples)

        if self.shuffle:
            np.random.shuffle(indices)

        num_of_batches = len(indices) // self.batch_size

        for i in range(num_of_batches):
            yield indices[i * self.batch_size : (i + 1) * self.batch_size]

        if len(indices) > num_of_batches * self.batch_size:
            yield indices[num_of_batches * self.batch_size :]

    def __len__(self):
        return self.num_of_samples  # // self.batch_size



from typing import Callable
from typing import Optional as O
from typing import Sequence
from typing import Union as U

import os
from h5py import File
import torch
from torch.utils.data import Dataset
from torchvision.transforms import Compose

from ..base.histo_dataset import HistoDataset
from ..base import data_readers, df_creators, df_manipulators
# TODO change to HistoDataset API


def pcam_patches(
    root: str,  # '/data/ldap/histopathologic/original_read_only/PCAM_extracted'
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    #   pre_transformation: O[U[Callable, Sequence[Callable]]] = None, # pretrans not possible yet
    subset: O[str] = "train",
) -> Dataset:
    """
    Creates the PCAM dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at https://humanunsupervised.github.io/humanunsupervised.com/pcam/pcam-cancer-detection.html

    Arguments:
        root: Absolute path to the dataset files.
        transformation: A callable or a list of callable transformations.
        subset: Which part of the dataset should be used.
                One of {"train", "validation", "test"}

    Returns:
        torch.utils.data.Dataset:
            The dataset that loads the pcam patches. You can use this
            as normal pytorch dataset. If you call it, the data will
            be returned as dictionary with structure:
            dict['images': torch.Tensor, 'labels': torch.Tensor].
    """
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        imgs_hdf5_filepath = root + "/camelyonpatch_level_2_split_train_x.h5"
        labels_hdf5_filepath = root + "/camelyonpatch_level_2_split_train_y.h5"
        return PCAM(
            imgs_hdf5_filepath=imgs_hdf5_filepath,
            labels_hdf5_filepath=labels_hdf5_filepath,
            transform=transformation,
        )
    elif subset.startswith("val"):
        imgs_hdf5_filepath = root + "/camelyonpatch_level_2_split_valid_x.h5"
        labels_hdf5_filepath = root + "/camelyonpatch_level_2_split_valid_y.h5"
        return PCAM(
            imgs_hdf5_filepath=imgs_hdf5_filepath,
            labels_hdf5_filepath=labels_hdf5_filepath,
            transform=transformation,
        )
    elif subset.startswith("te"):
        imgs_hdf5_filepath = root + "/camelyonpatch_level_2_split_test_x.h5"
        labels_hdf5_filepath = root + "/camelyonpatch_level_2_split_test_y.h5"
        return PCAM(
            imgs_hdf5_filepath=imgs_hdf5_filepath,
            labels_hdf5_filepath=labels_hdf5_filepath,
            transform=transformation,
        )
    # TODO use case not possible yet
    # elif subset == 'all':
    #     pass
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "validation", "test"].'
        )


class PCAM(Dataset):
    """
    Creates the PCAM dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at https://humanunsupervised.github.io/humanunsupervised.com/pcam/pcam-cancer-detection.html
    """

    def __init__(
        self,
        imgs_hdf5_filepath: str,
        labels_hdf5_filepath: str,
        imgs_key: str = "x",
        labels_key: str = "y",
        transform: O[Compose] = None,
    ):
        """
        Initializes dataset.
        """

        self.imgs_hdf5_filepath = imgs_hdf5_filepath
        self.labels_hdf5_filepath = labels_hdf5_filepath
        self.imgs_key = imgs_key
        self.labels_key = labels_key

        self.transform = transform
        self.loaded_images = None
        self.loaded_labels = None
        
    def __len__(self):
        """
        Returns length of dataset
        """
        with File(self.labels_hdf5_filepath, "r") as db:
            lens = len(db[self.labels_key])

        return lens

    def preload(self):
        with File(self.imgs_hdf5_filepath, "r") as db:
            self.loaded_images = db[self.imgs_key][::]
        with File(self.labels_hdf5_filepath, "r") as db:
            self.loaded_labels = db[self.labels_key][::]
        

    def get_feature_for_all_rows(self):
        labels = []
        with File(self.labels_hdf5_filepath, "r") as db:
            for idx in range(len(self)):
                label = db[self.labels_key][idx][0][0][0]  # TODO make it nice
                labels.append(label)
        return {"feature": labels}

    def __getitem__(self, idx: int) -> dict:
        """
        TODO
        """

        if self.loaded_images is None:
            # get images
            with File(self.imgs_hdf5_filepath, "r") as db:
                image = db[self.imgs_key][idx]
            # get labels
            with File(self.labels_hdf5_filepath, "r") as db:
                label = db[self.labels_key][idx][0][0][0]  # TODO make it nice
        else:
            image = self.loaded_images[idx]
            label = self.loaded_labels[idx][0][0][0]

        # transform data
        if self.transform:
            image = self.transform(image)

        return {"data": image, "feature": torch.as_tensor(label).type(torch.LongTensor)}

    
def pcam_patches_preprocessed(
    root: str,  # '/data/ldap/histopathologic/original_read_only/PCAM_extracted_images'
    transformation: O[U[Callable, Sequence[Callable]]] = None,
    pre_transformation: O[U[Callable, Sequence[Callable]]] = None,
    subset: O[str] = "train",
    seed: O[int] = None,
) -> HistoDataset:
    """
    Creates the midog dataset and returns it as type HistoDataset, which inherits from torch.utils.data.Dataset.

    Data and further information can be found at TODO

    Arguments:
        root: absolute path to the dataset files.
                transformation (Callable): A callable or a list of
                       callable transformations.
        pre_transformation (Callable): A callable or a list of
                       callable transformations. "pre_transformation"
                       is called before "transformation". The transformation
                       is called with a seed and will give always the same
                       result.
        subset: Which part of the dataset should be used.
                One of {"train", "valid", "test"}
        seed:   The seed that is used for the "pre_transformation" and the
                "transformation". If the seed is set, the "pre_transformation"
                will always use the same seed for a data row. If the seed is
                set, the "transformation" will use always the same seed
                depending on the call position.

    Returns:
        HistoDataset:
            The dataset that loads the bach patches. You can use this
            as normal pytorch dataset. If you call it, the data will
            be returned as dictionary.
    """
    if subset:
        subset = subset.lower()
    if subset.startswith("tr"):
        path_to_dataset = os.path.join(root, "train")
    elif subset.startswith("val"):
        path_to_dataset = os.path.join(root, "valid")
    elif subset.startswith("test"):
        path_to_dataset = os.path.join(root, "test")
    else:
        raise NotImplementedError(
            'The parameter "subset" needs to be one of ["train", "valid", "test"].'
        )

    path_to_csv = os.path.join(path_to_dataset, "features.csv")
    # create HistoDataset object and pass relevant attributes
    ds = HistoDataset(
        df_creators.CreateDFFromCSV(path_to_csv),
        path_to_dataset,
        data_readers="data/{id}.tiff",
        feature_readers=data_readers.ReadValueFromCSV(r"{feature}", encoded_values=["0", "1"]),
        seed=seed,
        pre_transfs=pre_transformation,
        da_transfs=transformation,
    )
    return ds